// CRC ZoNX CCrc
//

#ifndef __CRC_H__
#define __CRC_H__

#include <limits.h>

// new Zq̃}N` (VC++ ΍p)
#ifndef __NEW__
#define __NEW__
	#include <new>
	#include <windows.h>

	#define NEW(ptr, type)				if ((ptr = new (nothrow) type) == NULL) \
											throw std::bad_alloc()
	#define NEWARRAY(ptr, type, size)	if ((ptr = new type [size])    == NULL)	\
											throw std::bad_alloc()
	#define DEL(ptr)					delete ptr
	#define DELARRAY(ptr)				delete [] ptr
#endif	// __NEW__

/*
	CũRpCɂ́C
	CũNw𖳌Ƃ
*/
#ifndef _LIB
	#ifdef _DEBUG
		#pragma comment(lib, "datachkd.lib")
	#else
		#pragma comment(lib, "datachk.lib")
	#endif	// _DEBUG
#endif		// !_LIB

/*
	16 Bit CRC 

	CRC-16 The HDLC (CRC-CCITT)
	 x^16 + x^12 + x^5 + 1
	0001 0000 0010 0001 (0x1021) rbg
	1000 0100 0000 1000 (0x8408) rbgE

	CRC-16 (CRC-ANSI)
	 x^16 + x^15 + x^2 + 1
	1000 0000 0000 0101 (0x8005) rbg
	1010 0000 0000 0001 (0xa001) rbgE

	ȂCx^16  2Byte I[o[̂ߊ܂łȂ
*/
#define CCITT_CRC16L	0x1021U
#define CCITT_CRC16R	0x8408U
#define ANSI_CRC16L		0x8005U
#define	ANSI_CRC16R		0xa001U

/*
	32 Bit CRC 

	CRC-32
	 x^32 + x^26 + x^23 + x^22 + x^16 +
			   x^12 + x^11 + x^10 + x^8  + x^7  +
			   x^5  + x^4  + x^2  + x^1  + 1
	0000 0100 1100 0001 0001 1101 1011 0111 (0x04c11db7) rbg
	1110 1101 1011 1000 1000 0011 0010 0000 (0xedb88320) rbgE
*/
#define	CRC32L			0x04c11db7UL
#define CRC32R			0xedb88320UL

// CRC ZoNX
class CCrc
{
	protected:
		// CRC Qƃe[u쐬
		virtual void MakeCrcTable() throw (std::bad_alloc) = 0;

	public:
		// RXgN^
		CCrc(unsigned long int nCrcPoly, unsigned long int nCrcInit)
		{
			/*
				ʂɎgpĂ CRC ASY
				LHA Ŏgp
				 ANSI CRC-16 (rbgE)
				CRC l 0x0000U

				RAR(SFV) Ŏgp (16 Bit CRC ̏l)
				 CCITT CRC-16(rbgE)
				CRC l 0xffffU

				ZIP Ŏgp (32 Bit CRC ̏l)
				 CRC-32 (rbgE)
				CRC l 0xffffffffUL
				CRC l̎Zo 0xffffffffUL  rI_a (XOR) s
				 NOT(]) sƓӂł
			*/
			m_lpnCrcTable = NULL;
			m_nCrcPoly    = nCrcPoly;
			m_nCrcValue   = nCrcInit;
		}

		// fXgN^
		virtual ~CCrc()
		{
			if (m_lpnCrcTable)
				DELARRAY(m_lpnCrcTable);
		}

		// CRC vZ
		virtual void Calc(unsigned char cValue) = 0;

		// CRC ̒l擾
		virtual unsigned long int GetValue(unsigned long int nXorValue) = 0; 

	protected:
		unsigned long int m_nCrcValue;		// CRC ̒l

	protected:
		unsigned long int m_nCrcPoly;		// Ɏw肳ꂽ CRC ̐
		              int m_nBitShift;		// rbg̕ (0 = C1 = E)
		unsigned long int *m_lpnCrcTable;	// CRC Qƃe[uւ̃|C^
};

// 16 Bit CRC ZoNX
class CCrc16 : public CCrc
{
	protected:
		// 16 Bit CRC Qƃe[u쐬
		void MakeCrcTable() throw (std::bad_alloc);

	public:
		// RXgN^
		/*
			:
			@RXgN^ŗO

			@݂͓Tv̂
			@O𖳌Ă
		*/
		CCrc16(unsigned long int nCrcPoly = CCITT_CRC16R, 
			unsigned long int nCrcInit = 0x0000ffffU)
			: CCrc(nCrcPoly, nCrcInit)
		{
			switch (m_nCrcPoly)
			{
				case CCITT_CRC16L:
				case  ANSI_CRC16L:
					m_nBitShift = 0;
					break;

				case CCITT_CRC16R:
				case  ANSI_CRC16R:
					m_nBitShift = 1;
			}

			try
			{
				MakeCrcTable();
			}
			catch (...)
			{
			}
		}

		// fXgN^
		virtual ~CCrc16() {}

		// 16 Bit CRC vZ
		void Calc(unsigned char cValue);

		// 16 Bit CRC ̒l擾
		unsigned long int GetValue(unsigned long int nXorValue = 0x00000000UL)
		{
			return ((m_nCrcValue ^ nXorValue) & 0x0000ffffU);
		}
};

// 32 Bit CRC ZoNX
class CCrc32 : public CCrc
{
	protected:
		// 32 Bit CRC Qƃe[u쐬
		void MakeCrcTable() throw (std::bad_alloc);

	public:
		// RXgN^
		/*
			:
			@RXgN^ŗO

			@݂͓Tv̂
			@O𖳌Ă
		*/
		CCrc32(unsigned long int nCrcPoly = CRC32R, 
			unsigned long int nCrcInit = 0xffffffffUL)
			: CCrc(nCrcPoly, nCrcInit)
		{
			switch (m_nCrcPoly)
			{
				case CRC32L:
					m_nBitShift = 0;
					break;

				case CRC32R:
					m_nBitShift = 1;
			}

			try
			{
				MakeCrcTable();
			}
			catch (...)
			{
			}
		}

		// fXgN^
		virtual ~CCrc32() {}

		// 32 Bit CRC vZ
		void Calc(unsigned char cValue);

		// 32 Bit CRC ̒l擾
		unsigned long int GetValue(unsigned long int nXorValue = 0xffffffffUL)
		{
			return (m_nCrcValue ^ nXorValue);
		}
};

#endif	// !__CRC_H__
